#include <QTimer>
#include "miniaudioengine.h"
#include "dmxPersonality.h"

#define BIAS                1.0f
#define FILTER_ORDER        3

MiniAudioEngine::MiniAudioEngine() {}

void MiniAudioEngine::audioDataCallback(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount)
{
    ma_result result;

    result = ma_engine_read_pcm_frames((ma_engine*)pDevice->pUserData, pOutput, frameCount, NULL);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": error audio callback.";
    }
    (void)pInput;
}

void MiniAudioEngine::stopEngine()
{
    for (uint i = 0; i < m_mae.layersQty; i++) {
        if (m_mae.mediaLoaded[i])
            ma_sound_uninit(&m_mae.sounds[i]);
    }
    for (uint i = 0; i < m_mae.layersQty; i++) {
        ma_splitter_node_uninit(&m_mae.filters[i].input, NULL);
        ma_hpf_node_uninit(&m_mae.filters[i].hpf, NULL);
        ma_loshelf_node_uninit(&m_mae.filters[i].loshelf, NULL);
        ma_peak_node_uninit(&m_mae.filters[i].mLow, NULL);
        ma_peak_node_uninit(&m_mae.filters[i].mHigh, NULL);
        ma_hishelf_node_uninit(&m_mae.filters[i].hishelf, NULL);
        ma_splitter_node_uninit(&m_mae.filters[i].output, NULL);
    }
    for (uint i = 0; i < m_mae.audioDevicesQty; i++) {
        if (i > 0) {
            ma_writer_node_uninit(&m_mae.sendAuxNode[i], NULL);
            ma_pcm_rb_uninit(&m_mae.auxBuffers[i]);
        }
        ma_engine_uninit(&m_mae.engines[i]);
        ma_device_uninit(&m_mae.devices[i]);
    }
    ma_context_uninit(&m_mae.context);
    ma_resource_manager_uninit(&m_mae.resourceManager);
}

bool MiniAudioEngine::startEngine(uint layers, uint* audioDevicesId, uint audioDevicesQty)
{
    ma_result result;

    m_mae.layersQty = layers;
    m_mae.audioDevicesId = audioDevicesId;
    m_mae.audioDevicesQty = audioDevicesQty;
    for (uint i =0; i < m_mae.layersQty; i++) {
        m_mae.mediaLoaded[i] = MA_FALSE;
        m_mae.currentStatus[i].status = Status::Iddle;
        m_mae.currentStatus[i].pan = 128;
        m_mae.currentStatus[i].pitch = 128;
        m_mae.currentStatus[i].vol = 0.0f;
        m_mae.currentStatus[i].cursor = 0;
        m_mae.currentStatus[i].updated = MA_FALSE;
        m_mae.currentStatus[i].fadeIn = FADE_TIME;
        m_mae.currentStatus[i].fadeOut = FADE_TIME;
    }
    result = this->startContext();
    if (result != MA_SUCCESS) return false;
    result = this->getAllAudioDevices();
    if (result != MA_SUCCESS) return false;
    result = this->startDevices();
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed start audio devices." << endl;
        return false;
    }
    result = this->setNodeGraph();
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to set node graph." << endl;
        return false;
    }
    for (uint i = 0; i < m_mae.audioDevicesQty; i++) {
        result = ma_engine_start(&m_mae.engines[i]);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to start audio device" << m_mae.audioDevicesId[i] << endl;
            return false;
        }
    }
    return true;
}

ma_result MiniAudioEngine::createFilterBank(uint layer)
{
    ma_result result;
    ma_node_graph *ng = ma_engine_get_node_graph(&m_mae.engines[0]);
    ma_node *endpoint = ma_engine_get_endpoint(&m_mae.engines[0]);
    filterBank *fb    = &m_mae.filters[layer];

    ma_splitter_node_config splitterConfig = ma_splitter_node_config_init(CHANNELS);
    splitterConfig.outputBusCount= 3;
    result = ma_splitter_node_init(ng, &splitterConfig, NULL, &fb->input);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init input node." << endl;
        return result;
    }

    fb->hpfConfig = ma_hpf_node_config_init(CHANNELS, SAMPLE_RATE, 16, FILTER_ORDER);
    result = ma_hpf_node_init(ng, &fb->hpfConfig, NULL, &fb->hpf);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init high pass filter node." << endl;
        return result;
    }

    fb->loshelfConfig = ma_loshelf_node_config_init(CHANNELS, SAMPLE_RATE, 0.0f, 1.0f, 30);
    result = ma_loshelf_node_init(ng, &fb->loshelfConfig, NULL, &fb->loshelf);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init low pass filter node." << endl;
        return result;
    }

    fb->mLowConfig = ma_peak_node_config_init(CHANNELS, SAMPLE_RATE, 0.0, 4.0, 200); // double gainDB, double q, double frequency);
    result = ma_peak_node_init(ng, &fb->mLowConfig, NULL, &fb->mLow);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init peak low filter node." << endl;
        return result;
    }

    fb->mHighConfig = ma_peak_node_config_init(CHANNELS, SAMPLE_RATE, 0.0, 0.0, 600); // double gainDB, double q, double frequency);
    result = ma_peak_node_init(ng, &fb->mHighConfig, NULL, &fb->mHigh);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init peak high filter node." << endl;
        return result;
    }

    fb->hishelfConfig = ma_hishelf_node_config_init(CHANNELS, SAMPLE_RATE, 0.0f, 1.0f, 20000);
    result = ma_hishelf_node_init(ng, &fb->hishelfConfig, NULL, &fb->hishelf);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init hi shelf filter node." << endl;
        return result;
    }
    ma_vumeter_node_config vuc = ma_vumeter_node_config_init(CHANNELS, FORMAT, SAMPLE_RATE);
    ma_vumeter_node_init(ng, &vuc, NULL, &fb->vumeter);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init vumeter node." << endl;
        return result;
    }
    splitterConfig.outputBusCount = m_mae.audioDevicesQty;
    result = ma_splitter_node_init(ng, &splitterConfig, NULL, &fb->output);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init output node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->input, 0, &fb->hpf, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach input node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->input, 1, &fb->vumeter, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach bypass connection." << endl;
        return result;
    }
    ma_node_set_output_bus_volume(&fb->input, 1, 0.0f);
    result = ma_node_attach_output_bus(&fb->hpf, 0, &fb->loshelf, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach high pass pass filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->loshelf, 0, &fb->mLow, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach low shelf filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->mLow, 0, &fb->mHigh, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach low peaks filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->mHigh, 0, &fb->hishelf, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach high peaks filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->hishelf, 0, &fb->vumeter, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach high shelf filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->vumeter, 0, &fb->output, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach high shelf filter node." << endl;
        return result;
    }
    if (m_mae.audioDevicesQty == 1) {
        result = ma_node_attach_output_bus(&fb->output, 0, endpoint, 0);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to attach output to endpoint." << endl;
            return result;
        }
    } else {
        result = ma_node_attach_output_bus(&fb->output, 0, &m_mae.sendAuxNode[1], 0);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to attach output node to aux send 1." << endl;
            return result;
        }
        result = ma_node_attach_output_bus(&fb->output, 1, &m_mae.sendAuxNode[1], 1);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to attach output node to aux send 1." << endl;
            return result;
        }
        for (uint i = 2; i < m_mae.audioDevicesQty; i++) {
            result = ma_node_attach_output_bus(&fb->output, i, &m_mae.sendAuxNode[i], 1);
            if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to attach output node to aux send 1." << endl;
                return result;
            }
        }
    }
    return result;
}

ma_result MiniAudioEngine::setNodeGraph() {
    ma_result result = MA_SUCCESS;

    ma_node_graph *ng = ma_engine_get_node_graph(&m_mae.engines[0]);
    for (uint i = 1; i < m_mae.audioDevicesQty; i++) {
          size_t sizeInFrames = SAMPLE_RATE;
          result = ma_pcm_rb_init(FORMAT, CHANNELS, sizeInFrames, NULL, NULL, &m_mae.auxBuffers[i]);
            if (result != MA_SUCCESS) {
                printf("Failed to initialize ring buffer.\n");
                return result;
            }
            ma_silence_pcm_frames(m_mae.auxBuffers[i].rb.pBuffer, sizeInFrames, FORMAT, CHANNELS);
            ma_writer_node_config writerConfig = ma_writer_node_config_init(CHANNELS, SAMPLE_RATE * 5, &m_mae.auxBuffers[i]);
            result = ma_writer_node_init(ng, &writerConfig, NULL, &m_mae.sendAuxNode[i]);
            if (result != MA_SUCCESS) {
                cout << "ERROR " << result << ": Failed to init writer node." << endl;
                return result;
            }
            result = ma_node_attach_output_bus(&m_mae.sendAuxNode[i], 0, ma_engine_get_endpoint(&m_mae.engines[0]), 0);
            if (result != MA_SUCCESS) {
                cout << "ERROR " << result << ": Failed to attach writer node." << endl;
                return result;
            }
            result = ma_data_source_rb_init(&m_mae.dataSourceRB[i], &m_mae.auxBuffers[i]);
            if (result != MA_SUCCESS) {
                cout << "Error " << result << ": Failed to init data source ring buffer" << endl;
                return result;
            }
            ma_data_source_node_config dataSupplyNodeConfig = ma_data_source_node_config_init(&m_mae.dataSourceRB[i]);
            result = ma_data_source_node_init(ma_engine_get_node_graph(&m_mae.engines[i]), &dataSupplyNodeConfig, NULL, &m_mae.dataSupplyNode[i]);
            if (result != MA_SUCCESS) {
                cout << "Error " << result << ": Failed to init data source node" << endl;
                return result;
            }
            result = ma_node_attach_output_bus(&m_mae.dataSupplyNode[i], 0, ma_engine_get_endpoint(&m_mae.engines[i]), 0);
            if (result != MA_SUCCESS) {
                cout << "Error " << result << ": Failed to attach data source rb node" << endl;
                return result;
            }
    }
    for (uint i = 0; i < m_mae.layersQty; i++) {
        result = this->createFilterBank(i);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed creating filter bank." << endl;
        }
    }
    return (result);
}

ma_result MiniAudioEngine::startDevices()
{
    ma_result result = MA_SUCCESS;
    ma_device_config deviceConfig;
    ma_engine_config engineConfig;

    deviceConfig = ma_device_config_init(ma_device_type_duplex);
    deviceConfig.capture.format     = m_mae.resourceManager.config.decodedFormat;
    deviceConfig.capture.channels   = CHANNELS;
    deviceConfig.playback.channels  = CHANNELS;
    deviceConfig.capture.shareMode  = ma_share_mode_shared;
    deviceConfig.playback.format    = m_mae.resourceManager.config.decodedFormat;
    deviceConfig.sampleRate         = m_mae.resourceManager.config.decodedSampleRate;
    deviceConfig.dataCallback       = audioDataCallback;
    engineConfig                  = ma_engine_config_init();
    engineConfig.pResourceManager = &m_mae.resourceManager;
    engineConfig.defaultVolumeSmoothTimeInPCMFrames = SAMPLE_RATE / 500;
    engineConfig.noAutoStart      = MA_TRUE;

    for (uint internalId = 0; internalId < m_mae.audioDevicesQty; internalId++) {
        //deviceConfig.capture.pDeviceID  = &m_mae.pPlaybackDeviceInfos[m_mae.audioDevicesId[internalId]].id;
        deviceConfig.playback.pDeviceID = &m_mae.pPlaybackDeviceInfos[m_mae.audioDevicesId[internalId]].id;
        deviceConfig.pUserData          = &m_mae.engines[internalId];
        result = ma_device_init(&m_mae.context, &deviceConfig, &m_mae.devices[internalId]);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to initialize audio device " << m_mae.pPlaybackDeviceInfos[m_mae.audioDevicesId[internalId]].name << endl;
            return result;
        }
        engineConfig.pDevice = &m_mae.devices[internalId];
        result = ma_engine_init(&engineConfig, &m_mae.engines[internalId]);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to initialize audio engine" << endl;
            return result;
        }
        cout << "Initialized Audio Device. internalId: " << internalId << " systemId: " << m_mae.audioDevicesId[internalId] << " " << m_mae.pPlaybackDeviceInfos[m_mae.audioDevicesId[internalId]].name << endl;
    }
    return result;
}

ma_result MiniAudioEngine::startContext()
{
    ma_result result;

    ma_resource_manager_config resourceManagerConfig = ma_resource_manager_config_init();
    resourceManagerConfig.decodedFormat     = FORMAT;
    resourceManagerConfig.decodedChannels   = CHANNELS;
    resourceManagerConfig.decodedSampleRate = SAMPLE_RATE;
    resourceManagerConfig.jobThreadCount    = MAX_LAYERS;
    result = ma_resource_manager_init(&resourceManagerConfig, &m_mae.resourceManager);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to initialize audio resource manager." << endl;
        return result;
    }
    result = ma_context_init(NULL, 0, NULL, &m_mae.context);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to initialize audio context." << endl;
    }
    return result;
}

// enum all audio devices in system
ma_result MiniAudioEngine::getAllAudioDevices()
{
    ma_result result;

    result = ma_context_get_devices(&m_mae.context, &m_mae.pPlaybackDeviceInfos, &m_mae.playbackDeviceCount, NULL, NULL);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to enumerate playback devices." << endl;
        ma_context_uninit(&m_mae.context);
        return result;
    }
    cout << "Audio devices available:" << endl;
    for (ma_uint32 iAvailableDevice = 0; iAvailableDevice < m_mae.playbackDeviceCount; iAvailableDevice += 1) {
        cout << iAvailableDevice << " : " << m_mae.pPlaybackDeviceInfos[iAvailableDevice].name << endl;
    }
    return result;
}

char* MiniAudioEngine::getDeviceName(uint id)
{
    return m_mae.pPlaybackDeviceInfos[m_mae.audioDevicesId[id]].name;

}

ma_result MiniAudioEngine::loadMedia(int layer, char *file)
{
    ma_result result;

    if (!file || !*file)
        return MA_INVALID_ARGS;
    if (m_mae.mediaLoaded[layer] == MA_TRUE)
    {
        m_mae.mediaLoaded[layer] = MA_FALSE;
        ma_sound_set_volume(&m_mae.sounds[layer], 0.0f);
        ma_sound_stop(&m_mae.sounds[layer]);
        ma_sound_uninit(&m_mae.sounds[layer]);
    }
    ma_sound_config soundConfig = ma_sound_config_init();
    soundConfig = ma_sound_config_init();
    soundConfig.pFilePath = file;
    soundConfig.pInitialAttachment = &m_mae.filters[layer].input;
    soundConfig.initialAttachmentInputBusIndex = 0;
    soundConfig.channelsIn = 0;
    soundConfig.channelsOut = CHANNELS;
    soundConfig.flags = MA_SOUND_FLAG_NO_SPATIALIZATION | MA_SOUND_FLAG_NO_DEFAULT_ATTACHMENT | MA_SOUND_FLAG_STREAM; //| MA_SOUND_FLAG_NO_PITCH
    result = ma_sound_init_ex(&m_mae.engines[0], &soundConfig, &m_mae.sounds[layer]);
    if (result != MA_SUCCESS) {
        cout << "Error" << result << ": Failed to load file " << file << endl;
        return result;
    }
    m_mae.currentStatus[layer].media = file;
    m_mae.currentStatus[layer].updated = MA_TRUE;
    m_mae.mediaLoaded[layer] = MA_TRUE;
    this->refreshValues(layer);
    return result;
}

float MiniAudioEngine::getDuration(int layer)
{
    ma_result result;
    float ret;

    if (m_mae.mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_length_in_seconds(&m_mae.sounds[layer], &ret);
    if (result != MA_SUCCESS) {
        return result;
    }
    return (ret * 1000);
}

float MiniAudioEngine::getCursor(int layer)
{
    ma_result result;
    float ret = 0;

    if (m_mae.mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_cursor_in_seconds(&m_mae.sounds[layer], &ret);
    if (result != MA_SUCCESS)
    {
        cout << "Error" << result << ": Can not get cursor " << layer << endl;
        ret = MA_ERROR;
    }
    return ret;
}

ma_result MiniAudioEngine::printFormatInfo(int layer)
{
    ma_format format;
    ma_uint32 channels;
    ma_uint32 sampleRate;

    if (m_mae.mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    ma_result result = ma_sound_get_data_format(&m_mae.sounds[layer], \
                                    &format, &channels, &sampleRate, NULL, 0);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to get data format " << layer;
        cout << endl;
    } else {
        cout << "Layer:" << layer << " ";
        cout << m_mae.currentStatus[layer].media.toLatin1().data();
        cout << " samples/sec:" << sampleRate << " format:" <<  format;
        cout << " channels:" << channels << endl;
    }
    return result;
}

// Expects between 0 and 65535 vol value
void MiniAudioEngine::volChanged(int layer, int vol)
{
	volChanged(layer, vol, FADE_TIME, FADE_TIME);
}

// Expects between 0 and 65535 vol value and fade time in milliseconds
void MiniAudioEngine::volChanged(int layer, int vol, int fadeIn, int fadeOut)
{
    m_mae.currentStatus[layer].vol = vol;
    m_mae.currentStatus[layer].fadeIn = fadeIn;
	m_mae.currentStatus[layer].fadeOut = fadeOut;
    if (m_mae.mediaLoaded[layer] != MA_TRUE)
       return;
    if (m_mae.mediaLoaded[layer] == MA_FALSE && m_mae.currentStatus[layer].updated)
        return;
    float db = ((float)vol / 771.0f) - 85.0f;
    if (db <= -85.0f) {
        db = 0;
    } else
        db = ma_volume_db_to_linear(db);
    ma_sound_set_fade_in_milliseconds(&m_mae.sounds[layer], -1, db, fadeIn);
}

void MiniAudioEngine::panChanged(int layer, float value)
{
    float result;

    m_mae.currentStatus[layer].pan = value;
    if (m_mae.mediaLoaded[layer] != MA_TRUE)
        return;
    result = (value / 128.0) - 1.0;
    ma_sound_group_set_pan(&m_mae.sounds[layer], result);
}

void MiniAudioEngine::pitchChanged(int layer, float value)
{
    float pitch;

    m_mae.currentStatus[layer].pitch = value;
    if (m_mae.mediaLoaded[layer] != MA_TRUE)
        return;
    pitch = value / 128.0;
    ma_sound_group_set_pitch(&m_mae.sounds[layer], pitch);
}

ma_result MiniAudioEngine::playbackChanged(int layer, Status status)
{
    ma_result result = MA_SUCCESS;
    float db = 0;
    bool loop = false;
	bool playAtBegin = false;

	if (status == Status::Stopped)
		playAtBegin = true;	
    m_mae.currentStatus[layer].status = status;
    if (m_mae.mediaLoaded[layer] != MA_TRUE)
        return MA_DOES_NOT_EXIST;
    m_mae.currentStatus[layer].updated = false;
    switch (status) {
        case Status::Paused:
            result = ma_sound_stop_with_fade_in_milliseconds(&m_mae.sounds[layer], m_mae.currentStatus[layer].fadeOut);
            break;
        case Status::Stopped:
			ma_sound_stop_with_fade_in_milliseconds(&m_mae.sounds[layer], m_mae.currentStatus[layer].fadeOut);
            QTimer::singleShot(m_mae.currentStatus[layer].fadeOut, [this, layer] {
                this->seekToCursor(layer, m_mae.currentStatus[layer].cursor);
            });
            break;
        case Status::PlayingLoop:
            loop = true;
            if (m_mae.currentStatus[layer].cursor > 0) {
                result = this->seekToCursor(layer, m_mae.currentStatus[layer].cursor);
             }
        case Status::PlayingOnce:
        case Status::PlayingFolder:
        case Status::PlayingFolderLoop:
        case Status::PlayingFolderRandom:
			if (playAtBegin)
	            this->seekToCursor(layer, 0);
            ma_sound_set_looping(&m_mae.sounds[layer], loop);
            if (ma_sound_is_playing(&m_mae.sounds[layer])) break;
            ma_sound_set_stop_time_in_milliseconds(&m_mae.sounds[layer], ~(ma_uint64)0);
            db = (m_mae.currentStatus[layer].vol / 771.0f) - 85.0f;
            if (db <= -85.0f) db = 0;
            else db = ma_volume_db_to_linear(db);
            result = ma_sound_start(&m_mae.sounds[layer]);
            ma_sound_set_fade_in_milliseconds(&m_mae.sounds[layer], 0.000001f, 0.000000f, FADE_TIME);
            if (m_mae.currentStatus[layer].cursor > 0)
                usleep(FADE_TIME * 1500); // avoid glitch when seeking
            ma_sound_set_fade_in_milliseconds(&m_mae.sounds[layer], 0, db, m_mae.currentStatus[layer].fadeIn);
        default:
            break;
    }
    m_mae.currentStatus[layer].updated = true;
    return result;
}

ma_result MiniAudioEngine::setRangePoint(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;
    ma_uint64 end = 0, start;

    if (m_mae.mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    if (cursor == 0)
        start = 0;
    else {
        result = ma_sound_get_length_in_pcm_frames(&m_mae.sounds[layer], &end);
        if (result != MA_SUCCESS) { return result; }
        start = (cursor * end) / 65535;
    }
    result = ma_data_source_set_range_in_pcm_frames(&m_mae.sounds[layer].pDataSource, start, end);
    if (result != MA_SUCCESS)
        cout << "ERROR " << result << " :set range point" << endl;
    return (result);
}

ma_result MiniAudioEngine::setLoopPoint(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;
    ma_uint64 end = 0, start;

    if (m_mae.mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    if (cursor == 0)
        start = 0;
    else {
        result = ma_sound_get_length_in_pcm_frames(&m_mae.sounds[layer], &end);
        if (result != MA_SUCCESS) { return result; }
        start = (cursor * end) / 65535;
    }
    result = ma_data_source_set_loop_point_in_pcm_frames(&m_mae.sounds[layer].pDataSource, start + 1, end - 1);
    if (result != MA_SUCCESS)
        cout << "ERROR " << result << " :set loop point" << endl;
    return (result);
}

ma_result MiniAudioEngine::seekToCursor(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;
    ma_uint64 end = 0, start;

    if (m_mae.mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    if (cursor == 0)
        start = 0;
    else {
        result = ma_sound_get_length_in_pcm_frames(&m_mae.sounds[layer], &end);
        if (result != MA_SUCCESS) { return result; }
        start = (cursor * end) / 65535;
    }
    result = ma_sound_seek_to_pcm_frame(&m_mae.sounds[layer], start);
    return (result);
}

ma_result MiniAudioEngine::setCursor(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;

    m_mae.currentStatus[layer].cursor = cursor;
    result = this->seekToCursor(layer, cursor);
    return (result);
}

Status MiniAudioEngine::getStatus(int layer)
{
    return m_mae.currentStatus[layer].status;
}

void MiniAudioEngine::refreshValues(int layer)
{
    this->panChanged(layer, m_mae.currentStatus[layer].pan);
    this->pitchChanged(layer, m_mae.currentStatus[layer].pitch);
    this->playbackChanged(layer, m_mae.currentStatus[layer].status);
}

ma_result MiniAudioEngine::filterParamChanged(int layer, int channel, int value)
{
    ma_result result = MA_SUCCESS;
    filterBank *fb = &m_mae.filters[layer];

    if (channel == HP_FREQ) {
        fb->hpfConfig.hpf.cutoffFrequency = double((value * 1.31) + 16.0f); // 16 - 350
        result = ma_hpf_node_reinit(&fb->hpfConfig.hpf, &fb->hpf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set frecuency high pass filter node." << endl;
            return result;
        }
    } else if (channel == LOW_FREQ) {
        fb->loshelfConfig.loshelf.frequency = 30 + (value * 1.647); // 30 - 450
        result = ma_loshelf_node_reinit(&fb->loshelfConfig.loshelf, &fb->loshelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set frecuency low shelf filter node." << endl;
            return result;
        }
    } else if (channel == LOW_Q) {
        fb->loshelfConfig.loshelf.shelfSlope = (double)(value / 32.0f) + 0.1f; // 0.1 - 8
        result = ma_loshelf_node_reinit(&fb->loshelfConfig.loshelf, &fb->loshelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set Q low shelf filter node." << endl;
            return result;
        }
    } else if (channel == LOW_GAIN) {
        fb->loshelfConfig.loshelf.gainDB = (double)(value / 21.25f) - 6.023528412f;
        result = ma_loshelf_node_reinit(&fb->loshelfConfig.loshelf, &fb->loshelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set gain low shelf filter node." << endl;
            return result;
        }
    } else if (channel == MIDLOW_FREQ) {
        fb->mLowConfig.peak.frequency = 200 + (value * 9.019607843); // 200 - 450
        result = ma_peak_node_reinit(&fb->mLowConfig.peak, &fb->mLow);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set frecuency Mid Low pass filter node." << endl;
            return result;
        }
    } else if (channel == MIDLOW_Q) {
        fb->mLowConfig.peak.q = (double)( value / 64.0f) + 0.10; // 0.1 - 4
        result = ma_peak_node_reinit(&fb->mLowConfig.peak, &fb->mLow);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set Q Mid Low filter node." << endl;
            return result;
        }
    } else if (channel == MIDLOW_GAIN) {
        fb->mLowConfig.peak.gainDB = (double)(value / 7.0833333333333f) - 18.0f;
        result = ma_peak_node_reinit(&fb->mLowConfig.peak, &fb->mLow);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set gain Mid Low filter node." << endl;
            return result;
        }
    } else if (channel == MIDHIGH_FREQ) {
        fb->mHighConfig.peak.frequency = 600 + (value * 25.09803922); // 600 - 7000
        result = ma_peak_node_reinit(&fb->mHighConfig.peak, &fb->mHigh);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set frecuency Mid High filter node." << endl;
            return result;
        }
    } else if (channel == MIDHIGH_Q) {
        fb->mHighConfig.peak.q = (double)( value / 64.0f) + 0.10; // 0.1 - 4
        result = ma_peak_node_reinit(&fb->mHighConfig.peak, &fb->mHigh);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set Q Mid High filter node." << endl;
            return result;
        }
    } else if (channel == MIDHIGH_GAIN) {
        fb->mHighConfig.peak.gainDB = (double)(value / 7.0833333333333f) - 18.0f;
        result = ma_peak_node_reinit(&fb->mHighConfig.peak, &fb->mHigh);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set gain Mid High filter node." << endl;
            return result;
        }
    }  else if (channel == HIGH_FREQ) {
    fb->hishelfConfig.hishelf.frequency = 1500 + (value * 56.8627451); // 1500 - 16000
    result = ma_hishelf_node_reinit(&fb->hishelfConfig.hishelf, &fb->hishelf);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to frecuency high shelf filter node." << endl;
        return result;
    }
    } else if (channel == HIGH_Q) {
        fb->hishelfConfig.hishelf.shelfSlope = (double)( value / 32.0f) + 0.1f;
        result = ma_hishelf_node_reinit(&fb->hishelfConfig.hishelf, &fb->hishelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set Q high shelf filter node." << endl;
            return result;
        }
    } else if (channel == HIGH_GAIN) {
        fb->hishelfConfig.hishelf.gainDB = (double)(value / 21.25) - 6.023528412f;
        result = ma_hishelf_node_reinit(&fb->hishelfConfig.hishelf, &fb->hishelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set gain high shelf filter node." << endl;
            return result;
        }
    } else if (channel == SEND1) {
        float db = ((float)value / 3.0f) - 85.0f;
        if (db <= -85.0f) {
            db = 0;
        } else
            db = ma_volume_db_to_linear(db);
        ma_node_set_output_bus_volume(&fb->output, 0, db);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set Send 1 Volume." << endl;
            return result;
        }
    }  else if (channel == SEND2) {
        float db = ((float)value / 3.0f) - 85.0f;
        if (db <= -85.0f) {
            db = 0;
        } else
            db = ma_volume_db_to_linear(db);
        ma_node_set_output_bus_volume(&fb->output, 1, db);
       if (result != MA_SUCCESS) {
           cout << "ERROR " << result << ": Failed set Send 2 Volume." << endl;
       }
       return result;
   }
   return (result);
}

bool MiniAudioEngine::setBypass(int audioDevice, int layer, bool bypass)
{
    (void)audioDevice;
    filterBank *fb = &m_mae.filters[layer];

    if (bypass) {
        ma_node_set_output_bus_volume(&fb->input, 1, 1.0f);
        ma_node_set_output_bus_volume(&fb->input, 0, 0.0f);
    } else {
        ma_node_set_output_bus_volume(&fb->input, 1, 0.0f);
        ma_node_set_output_bus_volume(&fb->input, 0, 1.0f);
    }
    return true;
}
